from enum import Enum

from biolib.compute_node.webserver.webserver_types import ComputeNodeInfo
from biolib.typing_utils import TypedDict, Optional, List

from biolib.biolib_api_client.app_types import AppVersionOnJob, RemoteHost


class JobState(Enum):
    AWAITING_INPUT = 'awaiting_input'
    CLIENT_ABORTED = 'client_aborted'
    COMPLETED = 'completed'
    FAILED = 'failed'
    IN_PROGRESS = 'in_progress'


class _Job(TypedDict):
    app_version: AppVersionOnJob
    arguments_override_command: bool
    auth_token: str
    caller_job: Optional[str]
    created_at: str
    federated_job_uuid: Optional[str]
    public_id: str
    remote_hosts_with_warning: List[RemoteHost]
    state: str
    user_id: Optional[str]


# type optional keys with total=False
class Job(_Job, total=False):
    custom_compute_node_url: str


class JobDict(Job):
    pass


class CloudJob(TypedDict):
    public_id: str
    reserved_cpu_in_nano_shares: int
    reserved_memory_in_bytes: int
    max_runtime_in_seconds: int


class JobWrapper(TypedDict):
    access_token: str
    BASE_URL: str  # TODO: refactor this to lower case
    compute_node_info: Optional[ComputeNodeInfo]
    job: Job
    cloud_job: Optional[CloudJob]
    job_temporary_dir: Optional[str]



from biolib.typing_utils import TypedDict


class ComputeNodeInfo(TypedDict):
    auth_token: str
    ip_address: str
    public_id: str
    pybiolib_version: str


class ShutdownTimes(TypedDict):
    auto_shutdown_time_in_seconds: int


class WebserverConfig(TypedDict):
    base_url: str
    compute_node_info: ComputeNodeInfo
    ecr_region_name: str
    is_dev: bool
    s3_general_storage_bucket_name: str
    s3_lfs_bucket_name: str
    shutdown_times: ShutdownTimes



from biolib import api
from biolib.biolib_api_client import JobDict
from biolib.jobs.job_result import JobResult
from biolib.typing_utils import Optional


class Job:

    def __init__(self, uuid: str, auth_token: Optional[str] = None):
        self._uuid: str = uuid
        self._auth_token: Optional[str] = auth_token

        self._job_dict: Optional[JobDict] = None
        self._result: Optional[JobResult] = None

    def __str__(self) -> str:
        return self._uuid

    @property
    def result(self) -> JobResult:
        if not self._result:
            self._result = JobResult(job_uuid=self._uuid, job_auth_token=self._auth_token)

        return self._result

    def get_stdout(self) -> bytes:
        return self.result.get_stdout()

    def get_stderr(self) -> bytes:
        return self.result.get_stderr()

    def get_exit_code(self) -> int:
        return self.result.get_exit_code()

    def save_files(self, output_dir: str) -> None:
        self.result.save_files(output_dir=output_dir)

    def get_status(self) -> str:
        return self._get_job_dict(force_refetch=True)['state']

    def _get_job_dict(self, force_refetch: bool = False) -> JobDict:
        if not self._job_dict or force_refetch:
            job_dict: JobDict = api.client.get(
                url=f'/jobs/{self._uuid}/',
                headers={'Job-Auth-Token': self._auth_token} if self._auth_token else None,
            ).json()
            self._job_dict = job_dict

        return self._job_dict



from biolib.biolib_binary_format import BioLibBinaryFormatBasePackage
from biolib.biolib_binary_format.utils import IndexableBuffer, InMemoryIndexableBuffer, LazyLoadedFile
from biolib.biolib_binary_format.module_output import ModuleOutput
from biolib.typing_utils import TypedDict, List, Optional, Dict


class Metadata(TypedDict):
    version: int
    type: int
    stdout_length: int
    stderr_length: int
    files_info_length: int
    files_data_length: int
    exit_code: int


class UnencryptedModuleOutput(BioLibBinaryFormatBasePackage):
    _version = 1
    _type = 11
    _metadata_byte_lengths = dict(
        # Note: order is important
        version=1,
        type=1,
        stdout_length=8,
        stderr_length=8,
        files_info_length=8,
        files_data_length=8,
        exit_code=2,
    )
    _metadata_length = sum(_metadata_byte_lengths.values())
    _file_path_length_bytes = 4
    _file_data_length_bytes = 8

    def __init__(self, buffer: IndexableBuffer):
        super().__init__()
        self._buffer = buffer

        self._metadata: Optional[Metadata] = None
        self._stdout: Optional[bytes] = None
        self._stderr: Optional[bytes] = None
        self._files: Optional[List[LazyLoadedFile]] = None

    def get_exit_code(self) -> int:
        metadata = self._get_metadata()
        return metadata['exit_code']

    def get_stdout(self) -> bytes:
        if self._stdout is None:
            metadata = self._get_metadata()
            self._stdout = self._buffer.get_data(start=self._metadata_length, length=metadata['stdout_length'])

        return self._stdout

    def get_stderr(self) -> bytes:
        if self._stderr is None:
            metadata = self._get_metadata()
            self._stderr = self._buffer.get_data(
                start=self._metadata_length + metadata['stdout_length'],
                length=metadata['stderr_length'],
            )

        return self._stderr

    def get_files(self) -> List[LazyLoadedFile]:
        metadata = self._get_metadata()
        if self._files is None:
            self._files = []
            if metadata['files_info_length'] == 0:
                return self._files

            files_info_start = self._metadata_length + metadata['stdout_length'] + metadata['stderr_length']
            files_info_buffer = InMemoryIndexableBuffer(
                data=self._buffer.get_data(start=files_info_start, length=metadata['files_info_length'])
            )

            files_data_pointer = files_info_start + metadata['files_info_length']
            while files_info_buffer.pointer < len(files_info_buffer):
                path_length = files_info_buffer.get_data_with_pointer_as_int(self._file_path_length_bytes)
                path = files_info_buffer.get_data_with_pointer_as_string(path_length)
                data_length = files_info_buffer.get_data_with_pointer_as_int(self._file_data_length_bytes)

                data_start = files_data_pointer
                files_data_pointer += data_length

                self._files.append(LazyLoadedFile(path=path, buffer=self._buffer, start=data_start, length=data_length))

        return self._files

    def _get_metadata(self) -> Metadata:
        if self._metadata is None:
            metadata_buffer = InMemoryIndexableBuffer(self._buffer.get_data(start=0, length=self._metadata_length))

            partial_metadata = {}
            for field_name, field_length in self._metadata_byte_lengths.items():
                value = metadata_buffer.get_data_with_pointer_as_int(length=field_length)  # type: ignore
                if field_name == 'version' and value != UnencryptedModuleOutput._version:
                    raise Exception('Version does not match')

                if field_name == 'type' and value != UnencryptedModuleOutput._type:
                    raise Exception('Type does not match')

                partial_metadata[field_name] = value

            self._metadata = partial_metadata  # type: ignore

        return self._metadata  # type: ignore

    def convert_to_serialized_module_output(self) -> bytes:
        files_dict: Dict[str, bytes] = {}

        metadata = self._get_metadata()
        files_info_start = self._metadata_length + metadata['stdout_length'] + metadata['stderr_length']

        # Get all file info data at once
        files_info_buffer = InMemoryIndexableBuffer(data=self._buffer.get_data(
            start=files_info_start,
            length=metadata['files_info_length'],
        ))

        # Get all file data at once
        files_data_buffer = InMemoryIndexableBuffer(data=self._buffer.get_data(
            start=files_info_start + metadata['files_info_length'],
            length=metadata['files_data_length'],
        ))

        files_data_pointer = 0
        while files_info_buffer.pointer < len(files_info_buffer):
            path_length = files_info_buffer.get_data_with_pointer_as_int(self._file_path_length_bytes)
            path = files_info_buffer.get_data_with_pointer_as_string(path_length)
            data_length = files_info_buffer.get_data_with_pointer_as_int(self._file_data_length_bytes)

            data_start = files_data_pointer
            files_data_pointer += data_length

            files_dict[path] = files_data_buffer.get_data(start=data_start, length=data_length)

        return ModuleOutput().serialize(
            exit_code=self.get_exit_code(),
            files=files_dict,
            stderr=self.get_stderr(),
            stdout=self.get_stdout(),
        )

    @staticmethod
    def write_to_file_from_legacy_module_output_file(
            legacy_module_output_file_path: str,
            output_file_path: str,
    ) -> None:
        with open(legacy_module_output_file_path, mode='rb') as input_file, \
                open(output_file_path, mode='wb') as output_file:

            bbf_version = int.from_bytes(input_file.read(1), byteorder='big')
            if bbf_version != 1:
                raise Exception(f'Unsupported BioLib Binary Format version: Got {bbf_version} expected 1')

            package_type = int.from_bytes(input_file.read(1), byteorder='big')
            if package_type != 2:
                raise Exception(f'Unsupported BioLib Binary Format type: Got {package_type} expected 2')

            output_file.write(UnencryptedModuleOutput._version.to_bytes(1, 'big'))
            output_file.write(UnencryptedModuleOutput._type.to_bytes(1, 'big'))

            stdout_length_bytes = input_file.read(8)
            stderr_length_bytes = input_file.read(8)
            legacy_files_data_length = int.from_bytes(input_file.read(8), byteorder='big')
            exitcode_bytes = input_file.read(2)

            output_file.write(stdout_length_bytes + stderr_length_bytes)

            stdout_bytes = input_file.read(int.from_bytes(stdout_length_bytes, byteorder='big'))
            stderr_bytes = input_file.read(int.from_bytes(stderr_length_bytes, byteorder='big'))

            files_data_start = input_file.tell()
            files_data_end = files_data_start + legacy_files_data_length

            files_info = bytearray()
            files_data_size = 0

            while input_file.tell() < files_data_end:
                path_length_bytes = input_file.read(4)
                path_length = int.from_bytes(path_length_bytes, byteorder='big')
                data_length_bytes = input_file.read(8)
                data_length = int.from_bytes(data_length_bytes, byteorder='big')
                path_bytes = input_file.read(path_length)

                files_info.extend(path_length_bytes + path_bytes + data_length_bytes)
                files_data_size += data_length

                # Skip the file data
                input_file.seek(data_length, 1)

            output_file.write(len(files_info).to_bytes(8, byteorder='big'))
            output_file.write(files_data_size.to_bytes(8, byteorder='big'))
            output_file.write(exitcode_bytes)
            output_file.write(stdout_bytes)
            output_file.write(stderr_bytes)
            output_file.write(files_info)

            input_file.seek(files_data_start)
            while input_file.tell() < files_data_end:
                path_length_bytes = input_file.read(4)
                path_length = int.from_bytes(path_length_bytes, byteorder='big')
                data_length_bytes = input_file.read(8)
                data_length = int.from_bytes(data_length_bytes, byteorder='big')
                input_file.seek(path_length, 1)  # Skip path

                default_chunk_size = 4096
                data_left = data_length

                while data_left > 0:
                    data_chunk = input_file.read(min(default_chunk_size, data_left))
                    data_left -= len(data_chunk)
                    output_file.write(data_chunk)



# pylint: disable=unsubscriptable-object
import os
import subprocess
import time
from datetime import datetime
from socket import gethostbyname, gethostname
import requests

from biolib import utils
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional
from biolib.biolib_api_client import BiolibApiClient
from biolib.compute_node.webserver.webserver_types import WebserverConfig, ComputeNodeInfo, ShutdownTimes


class CloudUtils:
    _webserver_config: Optional[WebserverConfig] = None

    @staticmethod
    def initialize() -> None:
        logger_no_user_data.info('Reporting availability...')
        CloudUtils._report_availability()

    @staticmethod
    def get_webserver_config() -> WebserverConfig:
        if CloudUtils._webserver_config:
            return CloudUtils._webserver_config

        CloudUtils._webserver_config = WebserverConfig(
            compute_node_info=ComputeNodeInfo(
                auth_token=CloudUtils._get_environment_variable('BIOLIB_COMPUTE_NODE_AUTH_TOKEN'),
                ip_address=gethostbyname(gethostname()),
                public_id=CloudUtils._get_environment_variable('BIOLIB_COMPUTE_NODE_PUBLIC_ID'),
                pybiolib_version=utils.BIOLIB_PACKAGE_VERSION,
            ),
            base_url=CloudUtils._get_environment_variable('BIOLIB_BASE_URL'),
            ecr_region_name=CloudUtils._get_environment_variable('BIOLIB_ECR_REGION_NAME'),
            s3_general_storage_bucket_name=CloudUtils._get_environment_variable(
                'BIOLIB_S3_GENERAL_STORAGE_BUCKET_NAME'
            ),
            s3_lfs_bucket_name=CloudUtils._get_environment_variable('BIOLIB_S3_LFS_BUCKET_NAME'),
            is_dev=CloudUtils._get_environment_variable('BIOLIB_DEV').upper() == 'TRUE',
            shutdown_times=ShutdownTimes(
                auto_shutdown_time_in_seconds=CloudUtils._get_environment_variable_as_int(
                    'BIOLIB_CLOUD_AUTO_SHUTDOWN_TIME_IN_SECONDS'
                ),
            )
        )

        return CloudUtils._webserver_config

    @staticmethod
    def deregister_and_shutdown() -> None:
        logger_no_user_data.debug('Deregistering...')

        if utils.IS_RUNNING_IN_CLOUD:
            config = CloudUtils.get_webserver_config()
            try:
                response = requests.post(url=f'{config["base_url"]}/api/jobs/deregister/', json={
                    'auth_token': config["compute_node_info"]["auth_token"],
                    'public_id': config["compute_node_info"]["public_id"],
                })

                if not response.ok:
                    response_content = response.content.decode()
                    logger_no_user_data.error(
                        f'Failed to deregister got status {response.status_code} and response {response_content}'
                    )

            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.error(f'Failed to deregister got error: {error}')

            logger_no_user_data.debug('Waiting 10 seconds and shutting down...')
            # Sleep for 10 seconds to ensure logs are written
            time.sleep(10)
            logger_no_user_data.debug('Shutting down...')
            try:
                subprocess.run(['sudo', 'shutdown', 'now'], check=True)
            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.error(f'Failed to shutdown got error: {error}')

        else:
            logger_no_user_data.error("Not deregistering as environment is not cloud")

    @staticmethod
    def finish_cloud_job(cloud_job_id: str, system_exception_code: Optional[int]) -> None:
        if not cloud_job_id:
            logger_no_user_data.error(
                'Finish cloud job was called but no cloud job was supplied. '
                f'System exception code: {system_exception_code}'
            )
            return

        config = CloudUtils.get_webserver_config()
        try:
            requests.post(
                url=f'{config["base_url"]}/api/jobs/cloud/finish/',
                json={
                    'auth_token': config["compute_node_info"]["auth_token"],
                    'cloud_job_id': cloud_job_id,
                    'system_exception_code': system_exception_code,
                },
                timeout=5,
            )

            opt_error_string = f' with error code {system_exception_code}' if system_exception_code else ''
            logger_no_user_data.debug(
                f'Cloud Job "{cloud_job_id}" was reported as finished' + opt_error_string
            )

        except Exception as error:  # pylint: disable=broad-except
            logger_no_user_data.error(f'Could not finish cloud job with id: {cloud_job_id} got error: {error}')

    @staticmethod
    def _report_availability() -> None:
        try:
            config = CloudUtils.get_webserver_config()
            compute_node_info = config['compute_node_info']
            api_client = BiolibApiClient.get()
            logger_no_user_data.debug(
                f'Registering with {compute_node_info} to host {api_client.base_url} at {datetime.now()}'
            )

            response: Optional[requests.Response] = None
            max_retries = 5
            for retry_count in range(max_retries):
                try:
                    response = requests.post(f'{api_client.base_url}/api/jobs/report_available/',
                                             json=compute_node_info)
                    break
                except Exception as error:  # pylint: disable=broad-except
                    logger_no_user_data.error(f'Self-registering failed with error: {error}')
                    if retry_count < max_retries - 1:
                        seconds_to_sleep = 1
                        logger_no_user_data.info(f'Retrying self-registering in {seconds_to_sleep} seconds')
                        time.sleep(seconds_to_sleep)

            if not response:
                raise BioLibError('Failed to register. Max retry limit reached')

            if response.status_code != 201:
                raise Exception("Non 201 error code")

        except Exception as exception:  # pylint: disable=broad-except
            logger_no_user_data.error(f'Shutting down as self register failed due to: {exception}')
            if not utils.IS_DEV:
                CloudUtils.deregister_and_shutdown()

    @staticmethod
    def _get_environment_variable(key: str) -> str:
        value = os.environ.get(key)
        # Purposely loose falsy check (instead of `is not None`) as empty string should fail
        if not value:
            raise Exception(f'CloudUtils: Missing environment variable "{key}"')

        return value

    @staticmethod
    def _get_environment_variable_as_int(key: str) -> int:
        return int(CloudUtils._get_environment_variable(key))
